"""Async tasks for federated_content_connector."""
from logging import getLogger

from celery import shared_task
from opaque_keys.edx.keys import CourseKey

from federated_content_connector.course_metadata_importer import CourseMetadataImporter

LOGGER = getLogger(__name__)


@shared_task()
def import_course_metadata(courserun_keys):
    """
    Task to fetch course metadata for courseruns represented by `courserun_keys`.

    Arguments:
        courserun_keys (list): courserun keys
    """
    LOGGER.info(f'[FEDERATED_CONTENT_CONNECTOR] import_course_metadata task triggered: Keys: {courserun_keys}')
    courserun_locators = [CourseKey.from_string(courserun_key) for courserun_key in courserun_keys]
    CourseMetadataImporter.import_specific_courses_metadata(courserun_locators)



""""Management command to refresh course metadata."""

import logging
from datetime import datetime, timedelta

from django.core.management import BaseCommand
from django.db.models import Q
from django.utils.timezone import now
from opaque_keys.edx.locator import CourseLocator

from federated_content_connector.course_metadata_importer import CourseMetadataImporter
from federated_content_connector.models import CourseDetails, CourseDetailsImportStatus

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    """Management command to refresh course metadata"""

    help = "Refresh course metadata"

    def handle(self, *args, **options):
        self.refresh_courses_metadata()

    @classmethod
    def refresh_courses_metadata(cls):
        """Refresh courses updated after last refresh."""
        timestamp_format = CourseDetailsImportStatus.TIMESTAMP_FORMAT
        course_data_modified_timestamps = []

        last_successful_import_timestamp = CourseDetailsImportStatus.last_successful_import_timestamp()
        # This will be true only once in a lifetime
        if last_successful_import_timestamp is None:
            logger.info('[REFRESH_COURSE_METADATA] No previous timstamp found.')
            timestamp = now() - timedelta(hours=1)
            last_successful_import_timestamp = timestamp.strftime(timestamp_format)
            CourseDetailsImportStatus.save_last_successful_import_timestamp(last_successful_import_timestamp)

        logger.info(f'[REFRESH_COURSE_METADATA] Refresh Started. Timestamp: [{last_successful_import_timestamp}]')

        for courses in CourseMetadataImporter.courses(last_successful_import_timestamp):
            course_keys = [course['key'] for course in courses]
            course_data_modified_timestamps.extend([course['data_modified_timestamp'] for course in courses])

            logger.info(f'[REFRESH_COURSE_METADATA] Processing. Courses: [{course_keys}]')

            courserun_locators = cls.courseruns_to_update(course_keys)

            courserun_keys = list(map(str, courserun_locators))
            logger.info(f'[REFRESH_COURSE_METADATA] Updating. Courserun Kyes: [{courserun_keys}]')

            processed_courses_details = CourseMetadataImporter.process_courses_details(courserun_locators, courses)
            CourseMetadataImporter.store_courses_details(processed_courses_details)

            logger.info(f'[REFRESH_COURSE_METADATA] Processing Completed. Courses: [{course_keys}]')

        # Sort course timestamps in descending order and store first timestamp as last_successful_import_timestamp
        if course_data_modified_timestamps:
            logger.info(f'[REFRESH_COURSE_METADATA] All Course Timestamps: [{course_data_modified_timestamps}]')
            sorted_timestamps = sorted(
                course_data_modified_timestamps,
                key=lambda timestamp: datetime.strptime(timestamp, timestamp_format),
                reverse=True
            )
            next_timestamp = sorted_timestamps[0]
            CourseDetailsImportStatus.save_last_successful_import_timestamp(next_timestamp)

            logger.info(f'[REFRESH_COURSE_METADATA] Next Timestamp: [{next_timestamp}]')

        logger.info('[REFRESH_COURSE_METADATA] Refresh Completed.')

    @classmethod
    def courseruns_to_update(cls, course_keys):
        """Return a list of courserun locators"""
        # This is equivalent to `course-v1`
        namespace = CourseLocator.CANONICAL_NAMESPACE
        # Convert AA+AA101 to course-v1:AA+AA101
        namespaced_course_keys = map(lambda ck: f'{namespace}:{ck}', course_keys)

        qobjs = Q()
        for course_key in namespaced_course_keys:
            qobjs |= Q(id__istartswith=course_key)

        return list(CourseDetails.objects.filter(qobjs).values_list('id', flat=True))



# Generated by Django 3.2.19 on 2023-06-16 12:12

from django.db import migrations, models
import django_extensions.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ('federated_content_connector', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='CourseDetailsImportStatus',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('last_successful_import_at', models.DateTimeField(help_text='Timestamp of last data import')),
            ],
        ),
    ]



"""Course metadata importer."""

import logging
from urllib.parse import quote_plus, urlencode, urljoin

import backoff
from common.djangoapps.course_modes.models import CourseMode
from django.contrib.auth import get_user_model
from openedx.core.djangoapps.catalog.models import CatalogIntegration
from openedx.core.djangoapps.catalog.utils import get_catalog_api_base_url, get_catalog_api_client
from openedx.core.djangoapps.content.course_overviews.models import CourseOverview

from federated_content_connector.constants import BOOTCAMP_2U, EXEC_ED_COURSE_TYPE
from federated_content_connector.models import CourseDetails

BEST_MODE_ORDER = [
    CourseMode.VERIFIED,
    CourseMode.PROFESSIONAL,
    CourseMode.NO_ID_PROFESSIONAL_MODE,
    CourseMode.UNPAID_EXECUTIVE_EDUCATION,
    CourseMode.AUDIT,
]

logger = logging.getLogger(__name__)
User = get_user_model()


class CourseMetadataImporter:
    """
    Import course metadata from discovery.
    """

    @classmethod
    def get_api_client(cls):
        """
        Return discovery api client.
        """
        catalog_integration = CatalogIntegration.current()
        username = catalog_integration.service_username

        try:
            user = User.objects.get(username=username)
        except User.DoesNotExist:
            logger.exception(
                f'Failed to create API client. Service user {username} does not exist.'
            )
            raise

        return get_catalog_api_client(user)

    @classmethod
    def import_all_courses_metadata(cls):
        """
        Import course metadata for all courses.
        """
        logger.info('[COURSE_METADATA_IMPORTER] Course metadata import started for all courses.')

        all_active_courserun_locators = cls.courserun_locators_to_import()
        cls.import_courses_metadata(all_active_courserun_locators)

        logger.info('[COURSE_METADATA_IMPORTER] Course metadata import completed for all courses.')

    @classmethod
    def import_specific_courses_metadata(cls, courserun_locators):
        """
        Import course metadata for specific courses.

        Args:
            courserun_locators (list): list of courserun locator objects
        """
        logger.info(f'[COURSE_METADATA_IMPORTER] Course metadata import started for courses. {courserun_locators}')

        cls.import_courses_metadata(courserun_locators)

        logger.info(f'[COURSE_METADATA_IMPORTER] Course metadata import completed for courses. {courserun_locators}')

    @classmethod
    def import_courses_metadata(cls, courserun_locators):
        """
        Import course metadata for given course locators.

        Args:
            courserun_locators (list): list of courserun locator objects
        """
        logger.info('[COURSE_METADATA_IMPORTER] Course metadata import started.')

        client = cls.get_api_client()

        for courserun_locators_chunk in cls.chunks(courserun_locators):

            # convert course locator objects to courserun keys
            courserun_keys = list(map(str, courserun_locators_chunk))

            logger.info(f'[COURSE_METADATA_IMPORTER] Importing metadata. Courses: {courserun_keys}')

            course_details = cls.fetch_courses_details(client, courserun_locators_chunk, get_catalog_api_base_url())
            processed_courses_details = cls.process_courses_details(courserun_locators_chunk, course_details)
            cls.store_courses_details(processed_courses_details)

            logger.info(f'[COURSE_METADATA_IMPORTER] Import completed. Courses: {courserun_keys}')

        logger.info('[COURSE_METADATA_IMPORTER] Course metadata import completed for all courses.')

    @classmethod
    def courserun_locators_to_import(cls):
        """
        Construct list of all course locators for which we want to import data.
        """
        return list(CourseOverview.objects.all().values_list('id', flat=True))

    @classmethod
    def fetch_courses_details(cls, client, courserun_locators, api_base_url):
        """
        Fetch the course data from discovery using `/api/v1/courses` endpoint.
        """
        course_keys = [cls.construct_course_key(courserun_locator) for courserun_locator in courserun_locators]
        encoded_course_keys = ','.join(map(quote_plus, course_keys))

        logger.info(f'[COURSE_METADATA_IMPORTER] Fetching details from discovery. Courses {course_keys}.')
        api_url = urljoin(
            f"{api_base_url}/", f"courses/?limit=50&include_hidden_course_runs=1&keys={encoded_course_keys}"
        )
        response = cls.get_response_from_api(client, api_url)
        response.raise_for_status()
        courses_details = response.json()
        results = courses_details.get('results', [])

        # Find and log the course keys not found in course-discovery
        course_keys_in_response = [result.get('key') for result in results]
        courses_not_found = list(set(course_keys) - set(course_keys_in_response))
        if courses_not_found:
            logger.info(f'[COURSE_METADATA_IMPORTER] Courses not found in discovery. Courses: {courses_not_found}')

        return results

    @classmethod
    def courses(cls, timestamp):
        """Fetch courses updated since `timestamp`."""
        query_params = {
            'timestamp': timestamp,
            'limit': 50,
            'include_hidden_course_runs': 1,
        }
        client = cls.get_api_client()
        api_base_url = get_catalog_api_base_url()
        params = urlencode(query_params)
        api_url = urljoin(f"{api_base_url}/", f"courses/?{params}")
        results, next_url, total = cls.get_api_reponse(client, api_url)
        logger.info(f'[COURSE_METADATA_IMPORTER] Total Records are {total}')
        yield results

        while next_url:
            results, next_url, __ = cls.get_api_reponse(client, next_url)
            yield results

    @classmethod
    def get_api_reponse(cls, client, api_url):
        """Get response from API."""
        response = cls.get_response_from_api(client, api_url)
        response.raise_for_status()
        courses = response.json()
        results = courses.get('results', [])
        return results, courses.get('next'), courses.get('count')

    @classmethod
    def get_response_from_api(cls, client, api_url):
        """
        Call api endpoint and return response.
        """

        @backoff.on_exception(
            backoff.expo,
            Exception,
            max_tries=3,
            logger=logger,
        )
        def call_api():
            """
            Call api endpoint.
            """
            return client.get(api_url)

        return call_api()

    @classmethod
    def process_courses_details(cls, courserun_locators, courses_details):
        """
        Parse and extract the minimal data that we need.
        """
        log_prefix = 'COURSE_METADATA_IMPORTER'

        courses = {}
        for courserun_locator in courserun_locators:
            course_key = cls.construct_course_key(courserun_locator)
            courserun_key = str(courserun_locator)
            logger.info(f'[{log_prefix}] Process. CourserunKey: {courserun_key}, CourseKey: {course_key}')
            course_metadata = cls.find_attr(courses_details, 'key', course_key)
            if not course_metadata:
                logger.info(f'[COURSE_METADATA_IMPORTER] Metadata not found. CourseKey: {course_key}')
                continue

            course_type = course_metadata.get('course_type') or ''
            product_source = course_metadata.get('product_source') or ''
            if product_source:
                product_source = product_source.get('slug')

            enroll_by = start_date = end_date = None

            if course_type in (EXEC_ED_COURSE_TYPE, BOOTCAMP_2U):
                additional_metadata = course_metadata.get('additional_metadata')
                if additional_metadata:
                    enroll_by = additional_metadata.get('registration_deadline')
                    start_date = additional_metadata.get('start_date')
                    end_date = additional_metadata.get('end_date')
            else:
                course_run = cls.find_attr(course_metadata.get('course_runs'), 'key', courserun_key)
                if course_run:
                    seat = cls.find_best_mode_seat(course_run.get('seats'))
                    if seat:
                        enroll_by = seat.get('upgrade_deadline')
                    else:
                        logger.info(
                            f"[{log_prefix}] No Seat Found. Seats: {course_run.get('seats')}"
                        )
                    start_date = course_run.get('start')
                    end_date = course_run.get('end')
                else:
                    logger.info(
                        f'[{log_prefix}] Courserun not found. CourserunKey: {courserun_key}, CourseKey: {course_key}'
                    )

            course_data = {
                'course_type': course_type,
                'product_source': product_source,
                'enroll_by': enroll_by,
                'start_date': start_date,
                'end_date': end_date,
            }
            courses[courserun_key] = course_data

        return courses

    @classmethod
    def store_courses_details(cls, courses_details):
        """
        Store courses metadata in database.
        """
        for courserun_key, course_detail in courses_details.items():
            CourseDetails.objects.update_or_create(
                id=courserun_key,
                defaults=course_detail
            )

    @classmethod
    def find_best_mode_seat(cls, seats):
        """
        Find the seat by best course mode.
        """
        def sort_key(mode):
            """
            Assign a weight to the seat dictionary based on the position of its type in best moode order list.
            """
            mode_type = mode['type']
            if mode_type in BEST_MODE_ORDER:
                return len(BEST_MODE_ORDER) - BEST_MODE_ORDER.index(mode_type)
            else:
                return 0

        sorted_seats = sorted(seats, key=sort_key, reverse=True)
        if sorted_seats:
            return sorted_seats[0]

        return None

    @classmethod
    def chunks(cls, keys, chunk_size=50):
        """
        Yield chunks of size `chunk_size`.
        """
        for i in range(0, len(keys), chunk_size):
            yield keys[i:i + chunk_size]

    @staticmethod
    def construct_course_key(course_locator):
        """
        Construct course key from course run key.
        """
        return f'{course_locator.org}+{course_locator.course}'

    @classmethod
    def find_attr(cls, iterable, attr_name, attr_value):
        """
        Find value of an attribute from with in an iterable.
        """
        for item in iterable:
            if item[attr_name] == attr_value:
                return item

        return None



"""
URLs for federated_content_connector.
"""
from django.urls import re_path  # pylint: disable=unused-import
from django.views.generic import TemplateView  # pylint: disable=unused-import

urlpatterns = [
    # TODO: Fill in URL patterns and views here.
    # re_path(r'', TemplateView.as_view(template_name="federated_content_connector/base.html")),
]



"""
Database models for federated_content_connector.
"""
from django.db import models
from django_extensions.db.models import TimeStampedModel
from opaque_keys.edx.django.models import CourseKeyField


class CourseDetails(TimeStampedModel):
    """
    Model to store Course metadata.

    .. no_pii:
    """

    id = CourseKeyField(
        db_index=True,
        primary_key=True,
        max_length=255,
        help_text='Courserun key'
    )

    course_type = models.CharField(
        max_length=255,
        help_text='Type of course. For example Masters, Verified, Audit, executive-education-2u,  etc'
    )
    product_source = models.CharField(
        max_length=255,
        help_text='Tells about the origin of a course. For example, edx, 2u'
    )
    start_date = models.DateTimeField(
        null=True,
        help_text='The start date of the course.'
    )
    end_date = models.DateTimeField(
        null=True,
        help_text='The end date of the course.'
    )
    enroll_by = models.DateTimeField(
        null=True,
        help_text='The suggested deadline for enrollment.'
    )

    class Meta:
        """
        Meta class for CourseDetails.
        """

        app_label = 'federated_content_connector'


class CourseDetailsImportStatus(TimeStampedModel):
    """
    CourseDetails import status history.

    .. no_pii:
    """

    TIMESTAMP_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'

    last_successful_import_at = models.DateTimeField(help_text='Timestamp of last data import')

    @classmethod
    def last_successful_import_timestamp(cls):
        """
        Return `last_successful_import_at`.
        """
        last_import = cls.objects.first()
        if last_import:
            return last_import.last_successful_import_at.strftime(cls.TIMESTAMP_FORMAT)

        return None

    @classmethod
    def save_last_successful_import_timestamp(cls, timestamp):
        """
        Save `last_successful_import_at`.
        """
        last_import = cls.objects.first()
        if last_import:
            last_import.last_successful_import_at = timestamp
            last_import.save()
        else:
            cls.objects.create(last_successful_import_at=timestamp)

    class Meta:
        """
        Meta class for CourseDetailsImportStatus.
        """

        app_label = 'federated_content_connector'



"""Open edx Filters Pipeline for the federated content connector."""
from datetime import datetime

from django.conf import settings
from django.utils import timezone
from openedx.core.djangoapps.catalog.utils import get_course_data
from openedx_filters import PipelineStep
from pytz import utc

from federated_content_connector.constants import EXEC_ED_COURSE_TYPE, EXEC_ED_LANDING_PAGE, PRODUCT_SOURCE_2U
from federated_content_connector.models import CourseDetails


class CreateCustomUrlForCourseStep(PipelineStep):
    """
    Step that modifies the url for the course home page.

    Example usage:

    Add the following configurations to your configuration file:

        "OPEN_EDX_FILTERS_CONFIG": {
            "org.openedx.learning.course.homepage.url.creation.started.v1": {
                "fail_silently": False,
                "pipeline": [
                    "federated_content_connector.filters.pipeline.CreateCustomUrlForCourseStep"
                ]
            }
        }
    """

    def run_filter(self, course_key, course_home_url):  # pylint: disable=arguments-differ
        """
        Pipeline step that modifies the course home url for externally hosted courses
        """
        course_key_str = '{}+{}'.format(course_key.org, course_key.course)
        course_data = get_course_data(course_key_str, ['course_type', 'product_source'])
        if course_data:
            course_type = course_data.get('course_type')
            product_source = course_data.get('product_source')
            product_source_slug = product_source['slug'] if isinstance(product_source, dict) else product_source
            if course_type == EXEC_ED_COURSE_TYPE and product_source_slug == PRODUCT_SOURCE_2U:
                course_home_url = getattr(settings, 'EXEC_ED_LANDING_PAGE', EXEC_ED_LANDING_PAGE)

        return {'course_key': course_key, 'course_home_url': course_home_url}


class CreateApiRenderEnrollmentStep(PipelineStep):
    """
    Step that modifies the enrollment data for the course.

    Example usage:

    Add the following configurations to your configuration file:

        "OPEN_EDX_FILTERS_CONFIG": {
            "org.openedx.learning.home.enrollment.api.rendered.v1": {
                "fail_silently": False,
                "pipeline": [
                    "federated_content_connector.filters.pipeline.CreateApiRenderEnrollmentStep"
                ]
            }
        }
    """

    def run_filter(self, course_key, serialized_enrollment):  # pylint: disable=arguments-differ
        """
        Pipeline step that modifies the enrollment data for the course.
        """
        try:
            course_details = CourseDetails.objects.get(id=course_key)
            course_type = course_details.course_type
            product_source = course_details.product_source
            start_date = course_details.start_date
            if product_source == PRODUCT_SOURCE_2U and course_type == EXEC_ED_COURSE_TYPE:
                if start_date and start_date <= timezone.now():
                    serialized_enrollment['hasStarted'] = True
        except CourseDetails.DoesNotExist:
            pass

        return {'course_key': course_key, 'serialized_enrollment': serialized_enrollment}


class CreateApiRenderCourseRunStep(PipelineStep):
    """
    Step that modifies the courserun data for the course.

    Example usage:

    Add the following configurations to your configuration file:

        "OPEN_EDX_FILTERS_CONFIG": {
            "org.openedx.learning.home.courserun.api.rendered.started.v1": {
                "fail_silently": False,
                "pipeline": [
                    "federated_content_connector.filters.pipeline.CreateApiRenderCourseRunStep"
                ]
            }
        }
    """

    def run_filter(self, serialized_courserun):  # pylint: disable=arguments-differ
        """
        Pipeline step that modifies the courserun data for the course.
        """
        try:
            course_details = CourseDetails.objects.get(id=serialized_courserun.get('courseId'))
            course_type = course_details.course_type
            product_source = course_details.product_source
            homeUrl = serialized_courserun.get('homeUrl')
            start_date, end_date = course_details.start_date, course_details.end_date

            if product_source == PRODUCT_SOURCE_2U and course_type == EXEC_ED_COURSE_TYPE:
                now_utc = datetime.now(utc)
                serialized_courserun.update({
                    'startDate': start_date,
                    'endDate': end_date,
                    'isStarted': now_utc > start_date if start_date is not None else True,
                    'isArchived': now_utc > end_date if end_date is not None else False,
                    'resumeUrl': homeUrl
                })

        except CourseDetails.DoesNotExist:
            pass

        return {'serialized_courserun': serialized_courserun}



""""Management command to import course metadata for all existing courses."""

from django.core.management import BaseCommand

from federated_content_connector.course_metadata_importer import CourseMetadataImporter


class Command(BaseCommand):
    """Management command to import course metadata for all existing courses"""

    help = "Import course metadata for all existing courses"

    def handle(self, *args, **options):
        CourseMetadataImporter.import_all_courses_metadata()
